/*
* Test if there's facility for hierachiced priority inheritance
*/
#include <stdio.h>

/* Kernel includes. */
#include <FreeRTOS.h>
#include "task.h"
#include "queue.h"

#include "tools.h"

xQueueHandle xMutex1, xMutex2;
xTaskHandle mh1, mh2, mh3, norm;

static void MutexHolder1(void *);
static void MutexHolder2(void *);
static void MutexHolder3(void *);
static void Normal(void *);

void hierPiroInh(){
    xMutex1 = xQueueCreateMutex(queueQUEUE_TYPE_MUTEX);
    xMutex2 = xQueueCreateMutex(queueQUEUE_TYPE_MUTEX);

    xTaskCreate(MutexHolder1, (signed char *) "mh1", 1000, NULL, 1, & mh1);
    xTaskCreate(MutexHolder2, (signed char *) "mh2", 1000, NULL, 2, & mh2);
    xTaskCreate(MutexHolder3, (signed char *) "mh3", 1000, NULL, 4, & mh3);
    xTaskCreate(Normal, (signed char *) "Task1", 1000, NULL, 3, & norm);

    vTaskStartScheduler();
}

void MutexHolder1(void * xPara){
    long i;
    sPrint ("mh1 is running and try to take xMutex1!\n");

    if (xQueueTakeMutexRecursive(xMutex1, 0)){
        sPrint ("mh1 holding xMutex1!\n");
        vTaskResume(mh2);
        sPrint ("mh1(%d) is running!\n", uxTaskPriorityGet(mh1));
        vTaskResume(mh3);
        for ( ;; ){
            sPrint ("mh1(%d) is running!\n", uxTaskPriorityGet(mh1));
            for (i = 0; i <= 10000; i ++);
        }
    }else {
        sPrint ("mh1 failed to take xMutex1!\n");
    }
}

void MutexHolder2(void * xPara){
    vTaskSuspend(NULL);
    sPrint ("mh2 is running and try to take xMutex2!\n");

    if (xQueueTakeMutexRecursive(xMutex2, 0)){
        sPrint ("mh2 holding xMutex2!\n");
        sPrint ("mh2 try to take xMutex1!\n");
        for ( ;; ){
            if (xQueueTakeMutexRecursive(xMutex1, portMAX_DELAY)){
            
                sPrint ("mh2 holding xMutex1!\n");
            
            } else {

                sPrint ("mh2 failed to take xMutex1!\n");
            }

        }
    }
        
}

void MutexHolder3(void * xPara){
    vTaskSuspend(NULL);
    sPrint ("mh3 is running and try to take xMutex2!\n");
    vTaskResume(norm);
    for ( ;; ) {
        if (xQueueTakeMutexRecursive(xMutex2, portMAX_DELAY)){
            sPrint ("mh3 holding xMutex2!\n");
        }
    }

}

void Normal(void * xPara){
    long i;
    vTaskSuspend(NULL);
    for ( ;; ){
        sPrint ("norm is running\n");
        sPrint ("p(mh1) = %d, p(mh2) = %d, p(mh3) = %d\n",
            uxTaskPriorityGet(mh1),
            uxTaskPriorityGet(mh2),
            uxTaskPriorityGet(mh3));
        for (i = 0; i <= 1000000; i++);
    }
}