/*
* This scenario shows that the priority modification has no efficient to the order of
* waiting tasks.
*
* Initial:
*		Task1(2) would direct send items to the xQueue;
*		Task2(3) would be blocked for 1 tick, then send items to the xQueue;
*		Task3(2) would be blocked for 2 ticks, then send items to the xQueue;
*		Task4(1) would be blocked for 3 ticks, then send items to the xQueue;
*		Task5(0) recieve an item from xQueue every 20 ticks. Decrease the
*			priority of Task2 to 1 and increase the priority of Task4 from 1
*			to 4.
*
* Befor Task5 Recieve items from xQueue, Task1-4 would be blocked in order
* Task2(3) -> Task1(2) -> Task3(2) -> Task4(1), sorting by priority. When Task5 start
* to recieve items from xQueue, The priority of Task2 and Task4 would be changed.
* Meanwhile, Task1-4 would be released in old order. Once tasks released from the
* xQueue and sent the item to the xQueue, it would be blocked and insert to event list
* again sorted by new priority. Sorting only happened when tasks insert to the event
* list.
* 
*/
#include <stdio.h>

/* Kernel includes. */
#include <FreeRTOS.h>
#include "task.h"
#include "queue.h"
#include "tools.h"

static void TaskS(void *xPara);
static void TaskR(void *xPara);
xTaskHandle tar, tar1;
xQueueHandle xQueue;


void waitingList(void){
	xQueue = xQueueCreate(10, sizeof(long));

	if (xQueue != NULL){
		xTaskCreate(TaskS, (signed char *) "Task1", 1000, (void *) 0, 2, NULL);
		xTaskCreate(TaskS, (signed char *) "Task2", 1000, (void *) 1, 3, & tar);
		xTaskCreate(TaskS, (signed char *) "Task3", 1000, (void *) 2, 2, NULL);
		xTaskCreate(TaskS, (signed char *) "Task4", 1000, (void *) 3, 1, & tar1);

		xTaskCreate(TaskR, (signed char *) "Task5", 1000, NULL, 0, NULL);
	}
	vTaskStartScheduler();
}

static void TaskS(void * xPara){
	long d, v, oldTick;
	char * name;
	xTaskHandle self;
	d = (int) xPara;
	
	oldTick = 0;

	for( ;; ){
//		if (xTaskGetTickCount() != oldTick){
			if (oldTick < 4)
				vTaskDelay(d);
			oldTick = xTaskGetTickCount();
			v = (long) uxTaskPriorityGet(NULL);
			self = xTaskGetCurrentTaskHandle();
			name = getTaskName(self);
			
			sPrint("tick: %d ***(%s try to send: %d)**\n", oldTick, name, v);
				
			xQueueSend(xQueue, & v, portMAX_DELAY);
			sPrint("*****%s sent: %d \n", name, v);

//		}
	}
}

static void TaskR(void * xPara){
	long v;
	
	for( ;; ){
		vTaskDelay(20);
	
		vTaskPrioritySet(tar, 1);
		vTaskPrioritySet(tar1, 4);

		if (xQueuePeek(xQueue, &v, portMAX_DELAY) == pdPASS){
			sPrint("---%s recieved: %d---\n", getTaskName(xTaskGetCurrentTaskHandle()), v);
			xQueueReceive(xQueue, &v, portMAX_DELAY);
		}
		else
			sPrint("invalid\n");
	}
}